from ontobio.io.gafparser import GafParser
from ontobio.io.ontol_renderers import GraphRenderer
from ontobio.assoc_factory import AssociationSetFactory

POMBASE = "tests/resources/truncated-pombase.gaf"

def test_subont():
    from ontobio.ontol_factory import OntologyFactory
    ofa = OntologyFactory()
    ont = ofa.create('go')
    f = AssociationSetFactory()
    aset = f.create(ontology=ont, file=POMBASE, fmt='gaf')

    ont = aset.subontology(minimal=False)
    
    print("ONT = {} N={}".format(ont, len(ont.nodes())))
    w = GraphRenderer.create('obo')
    print(w.render(ont))

    



"""
Classes for representing ontologies backed by SciGraph endpoint

E.g.
https://scigraph-ontology.monarchinitiative.org/scigraph/docs/#!/graph/getNeighbors
"""

import networkx as nx
import logging
import ontobio.ontol
import requests
from ontobio.ontol import Ontology

class RemoteScigraphOntology(Ontology):
    """
    ontology backed by SciGraph endpoint
    """

    def __init__(self, url=None):
        if url is not None:
            self.url_prefix = url
        else:
            self.url_prefix = "http://scigraph-data.monarchinitiative.org/scigraph/"
        return
    
    # Internal wrapper onto requests API
    def get_response(self, path="", q=None, format=None, **params):
        url = self.url_prefix + path;
        if q is not None:
            url += "/" +q;
        if format is not None:
            url = url  + "." + format;
        r = requests.get(url, params=params)
        return r

    def neighbors(self, id=None, **params):
        """
        Get neighbors of a node

        parameters are directly passed through to SciGraph: e.g. depth, relationshipType
        """
        response = self.get_response("graph/neighbors", id, "json", **params)
        # TODO: should return ids?
        return response.json()
    
    def extract_subset(self, subset):
        pass
    
    def resolve_names(self, names, is_remote=False, **args):
        ## TODO
        return names
    
    def subgraph(self, nodes=[]):
        return self.get_graph().subgraph(nodes)

    



from ontobio.io.gafparser import GafParser
from ontobio.assoc_factory import AssociationSetFactory

URL = "ftp://ftp.rgd.mcw.edu/pub/ontology/annotated_rgd_objects_by_ontology/homo_genes_pw"

def test_skim():
    p = GafParser()
    results = p.skim(URL)
    print(str(results))

    
def test_parse():
    p = GafParser()
    results = p.parse(URL)
    for r in results:
        print(str(r))
    



"""
A module for representing simple graph-oriented views of an ontology

See also:

 - ontol_factory.py

"""

import networkx as nx
import logging
import re

class Ontology():
    """An object that represents a basic graph-oriented view over an ontology.

    The ontology may be represented in memory, or it may be located
    remotely. See subclasses for details.

    The default implementation is an in-memory wrapper onto the python networkx library

    """

    def __init__(self, handle=None, graph=None, xref_graph=None, payload=None, graphdoc=None):
        """
        initializes based on an ontology name.

        **Note**: do not call this directly, use OntologyFactory instead
        """
        self.handle = handle

        # networkx object
        self.graph = graph
        self.xref_graph = xref_graph

        # obograph
        self.graphdoc = graphdoc

        self.all_logical_definitions = []
        
        # alternatively accept a payload object
        if payload is not None:
            self.graph = payload.get('graph')
            self.xref_graph = payload.get('xref_graph')
            self.graphdoc = payload.get('graphdoc')
            self.all_logical_definitions = payload.get('logical_definitions')
        

    def get_graph(self):
        """
        Return a networkx graph for the whole ontology.

        Note: Only implemented for *eager* implementations 

        Return
        ------
        nx.MultiDiGraph
            A networkx MultiDiGraph object representing the complete ontology
        """
        return self.graph

    # consider caching
    def get_filtered_graph(self, relations=None, prefix=None):
        """
        Returns a networkx graph for the whole ontology, for a subset of relations

        Only implemented for eager methods.

        Implementation notes: currently this is not cached

        Arguments
        ---------
          - relations : list
             list of object property IDs, e.g. subClassOf, BFO:0000050. If empty, uses all.
          - prefix : String
             if specified, create a subgraph using only classes with this prefix, e.g. ENVO, PATO, GO
        Return
        ------
        nx.MultiDiGraph
            A networkx MultiDiGraph object representing the filtered ontology
        """
        # default method - wrap get_graph
        srcg = self.get_graph()
        if prefix is not None:
            srcg = srcg.subgraph([n for n in srcg.nodes() if n.startswith(prefix+":")])
        if relations is None:
            logging.info("No filtering on "+str(self))
            return srcg
        logging.info("Filtering {} for {}".format(self, relations))
        g = nx.MultiDiGraph()
    
        logging.info("copying nodes")
        for n,d in srcg.nodes_iter(data=True):
            g.add_node(n, attr_dict=d)

        logging.info("copying edges")
        num_edges = 0
        for x,y,d in srcg.edges_iter(data=True):
            if d['pred'] in relations:
                num_edges += 1
                g.add_edge(x,y,attr_dict=d)
        logging.info("Filtered edges: {}".format(num_edges))
        return g

    def merge(self, ontologies):
        """
        TODO: incomplete
        """
        for ont in ontologies:
            g = self.get_graph()
            xg = ont.get_graph()
            for n in xg.nodes():
                g.add_node(n)
            for (o,s,m) in xg.edges(data=True):
                g.add_edge(o,s,attr_dict=m)
            
        
    def subgraph(self, nodes=[]):
        """
        Return an induced subgraph

        By default this wraps networkx subgraph,
        but this may be overridden in specific implementations
        """
        return self.get_graph().subgraph(nodes)

    def subontology(self, nodes=None, minimal=False, relations=None):
        """
        Return a new ontology that is an extract of this one

        Arguments:

        - Nodes: list

            list of node IDs to include in subontology. If None, all are used

        - Relations: list

            list of relation IDs to include in subontology. If None, all are used

        """
        g = self.get_graph()
        if nodes is not None:
            g = g.subgraph(nodes)
        if minimal:
            from ontobio.slimmer import get_minimal_subgraph
            g = get_minimal_subgraph(self.get_graph(), nodes)
            
        ont = Ontology(graph=g) # TODO - add metadata
        if relations is not None:
            g = ont.get_filtered_graph(relations)
            ont = Ontology(graph=g)
        return ont

    def create_slim_mapping(self, subset=None, subset_nodes=None, relations=None, disable_checks=False):
        """
        Create a dictionary that maps between all nodes in an ontology to a subset

        Arguments
        ---------
        ont : `Ontology`
            Complete ontology to be mapped. Assumed pre-filtered for relationship types
        subset : str
            Name of subset to map to, e.g. goslim_generic
        nodes : list
            If no named subset provided, subset is passed in as list of node ids
        relations : list
            List of relations to filter on
        disable_checks: bool
            Unless this is set, this will prevent a mapping being generated with non-standard relations.
            The motivation here is that the ontology graph may include relations that it is inappropriate to
            propagate gene products over, e.g. transports, has-part
    
        Return
        ------
        dict
            maps all nodes in ont to one or more non-redundant nodes in subset

        Raises
        ------
        ValueError
            if the subset is empty
        """
        if subset is not None:
            subset_nodes = self.extract_subset(subset)
            logging.info("Extracting subset: {} -> {}".format(subset, subset_nodes))
        
        if subset_nodes is None or len(subset_nodes) == 0:
            raise ValueError("subset nodes is blank")
        subset_nodes = set(subset_nodes)
        logging.debug("SUBSET: {}".format(subset_nodes))

        # Use a sub-ontology for mapping
        subont = self
        if relations is not None:
            subont = self.subontology(relations=relations)
            
        if not disable_checks:
            for r in subont.relations_used():
                if r != 'subClassOf' and r != 'BFO:0000050' and r != 'subPropertyOf':
                    raise ValueError("Not safe to propagate over a graph with edge type: {}".format(r))
        
        m = {}
        for n in subont.nodes():
            ancs = subont.ancestors(n, reflexive=True)
            #logging.info("  M: {} -> {}".format(n, ancs))
            ancs_in_subset = subset_nodes.intersection(ancs)
            m[n] = list(ancs_in_subset)
        return m

    def filter_redundant(self, ids):
        """
        Return all non-redundant ids from a list
        """
        sids = set(ids)
        for id in ids:
            sids = sids.difference(self.ancestors(id, reflexive=False))
        return sids
    
    def extract_subset(self, subset):
        """
        Return all nodes in a subset.
    
        We assume the oboInOwl encoding of subsets, and subset IDs are IRIs, or IR fragments
        """
        pass

    def prefixes(self):
        """
        list all prefixes used
        """
        pset = set()
        for n in self.nodes():
            pfx = self.prefix(n)
            if pfx is not None:
                pset.add(pfx)
        return list(pset)

    def prefix(self, nid):
        """
        Return prefix for a node
        """
        parts = nid.split(":")
        if len(parts) > 1:
            return parts[0]
        else:
            return None
    
    def nodes(self):
        """
        Return all nodes in ontology

        Wraps networkx by default
        """
        return self.get_graph().nodes()

    def node(self, id):
        """
        Return a node with a given ID

        Wraps networkx by default
        """
        return self.get_graph().node[id]

    def relations_used(self):
        """
        Return list of all relations used to connect edges
        """
        g = self.get_graph()
        types = set()
        for x,y,d in g.edges_iter(data=True):
            types.add(d['pred'])
        return list(types)
    
    def neighbors(self, node, relations=None):
        return self.parents(node, relations=relations) + self.children(node, relations=relations)
        
    def parents(self, node, relations=None):
        """
        Return all direct parents of specified node.

        Wraps networkx by default.

        Arguments
        ---------

        node: string

           identifier for node in ontology

        relations: list of strings

           list of relation (object property) IDs used to filter

        """
        g = None
        if relations is None:
            g = self.get_graph()
        else:
            # TODO: make this more efficient
            g = self.get_filtered_graph(relations)
        if node in g:
            return g.predecessors(node)
        else:
            return []
    
    def children(self, node, relations=None):
        """
        Return all direct children of specified node.

        Wraps networkx by default.

        Arguments
        ---------

        node: string

           identifier for node in ontology

        relations: list of strings

           list of relation (object property) IDs used to filter

        """
        g = None
        if relations is None:
            g = self.get_graph()
        else:
            g = self.get_filtered_graph(relations)
        if node in g:
            return g.successors(node)
        else:
            return []
    
    def ancestors(self, node, relations=None, reflexive=False):
        """Return all ancestors of specified node.

        The default implementation is to use networkx, but some
        implementations of the Ontology class may use a database or
        service backed implementation, for large graphs.

        Arguments
        ---------
        node : str
            identifier for node in ontology
        reflexive : bool
            if true, return query node in graph
        relations : list
             relation (object property) IDs used to filter

        Returns
        -------
        list[str]
            ancestor node IDs

        """
        if reflexive:
            ancs = self.ancestors(node, relations, reflexive=False)
            ancs.add(node)
            return ancs
            
        g = None
        if relations is None:
            g = self.get_graph()
        else:
            g = self.get_filtered_graph(relations)
        if node in g:
            return nx.ancestors(g, node)
        else:
            return []

    def descendants(self, node, relations=None, reflexive=False):
        """
        Returns all ancestors of specified node.

        The default implementation is to use networkx, but some
        implementations of the Ontology class may use a database or
        service backed implementation, for large graphs.


        Arguments
        ---------
        node : str
            identifier for node in ontology
        reflexive : bool
            if true, return query node in graph
        relations : list
             relation (object property) IDs used to filter

        Returns
        -------
        list[str]
            descendant node IDs
        """
        if reflexive:
            ancs = self.ancestors(node, relations, reflexive=False)
            return ancs + [node]
        g = None
        if relations is None:
            g = self.get_graph()
        else:
            g = self.get_filtered_graph(relations)
        if node in g:
            return nx.descendants(g, node)
        else:
            return []

    def traverse_nodes(self, qids, up=True, down=False, **args):
        """
        Traverse (optionally) up and (optionally) down from an input set of nodes

        Arguments
        ---------
        qids : list[str]
            list of seed node IDs to start from
        up : bool
            if True, include ancestors
        down : bool
            if True, include descendants
        relations : list[str]
            list of relations used to filter

        Return
        ------
        list[str]
            nodes reachable from qids
        """
        g = self.get_filtered_graph(**args)
        nodes = set()
        for id in qids:
            # reflexive - always add self
            nodes.add(id)
            if down:
                nodes.update(nx.descendants(g, id))
            if up:
                nodes.update(nx.ancestors(g, id))
        return nodes
                
    def get_roots(self, relations=None, prefix=None):
        """
        Get all nodes that lack parents

        Arguments
        ---------
        relations : list[str]
            list of relations used to filter
        prefix : str
            E.g. GO. Exclude nodes that lack this prefix when testing parentage
        """
        g = self.get_filtered_graph(relations=relations, prefix=prefix)
        # note: we also eliminate any singletons, which includes obsolete classes
        roots = [n for n in g.nodes() if len(g.predecessors(n)) == 0 and len(g.successors(n)) > 0]
        return roots
        
    def get_level(self, level, relations=None, **args):
        """
        Get all nodes at a particular level

        Arguments
        ---------
        relations : list[str]
            list of relations used to filter
        """
        g = self.get_filtered_graph(relations)
        nodes = self.get_roots(relations=relations, **args)
        for i in range(level):
            logging.info(" ITERATING TO LEVEL: {} NODES: {}".format(i, nodes))
            nodes = [c for n in nodes
                     for c in g.successors(n)]
        logging.info(" FINAL: {}".format(nodes))
        return nodes

    def parent_index(self, relations=None):
        """
        Returns a mapping of nodes to all direct parents

        Arguments
        ---------
        relations : list[str]
            list of relations used to filter

        Returns:
        list
            list of lists [[CLASS_1, PARENT_1,1, ..., PARENT_1,N], [CLASS_2, PARENT_2,1, PARENT_2,2, ... ] ... ]
        """
        g = None
        if relations is None:
            g = self.get_graph()
        else:
            g = self.get_filtered_graph(relations)
        l = []
        for n in g:
            l.append([n] ++ g.predecessors(b))
        return l
        
    def logical_definitions(self, nid):
        """
        Retrieves logical definitions for a class id

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried

        Returns
        -------
        LogicalDefinition
        """
        ldefs = self.all_logical_definitions
        if ldefs is not None:
            #print("TESTING: {} AGAINST LD: {}".format(nid, str(ldefs)))
            return [x for x in ldefs if x.class_id == nid]
        else:
            return []

    def synonyms(self, nid, include_label=False):
        """
        Retrieves synonym objects for a class

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        include_label : bool
            If True, include label/names as Synonym objects

        Returns
        -------
        list[Synonym]
            :class:`Synonym` objects
        """
        n = self.node(nid)
        syns = []
        if 'meta' in n:
            meta = n['meta']
            if 'synonyms' in meta:
                for obj in meta['synonyms']:
                    syns.append(Synonym(nid, **obj))
        if include_label:
            syns.append(Synonym(nid, val=self.label(nid), pred='label'))
        return syns

    def all_synonyms(self, include_label=False):
        """
        Retrieves all synonyms

        Arguments
        ---------
        include_label : bool
            If True, include label/names as Synonym objects

        Returns
        -------
        list[Synonym]
            :class:`Synonym` objects
        """
        syns = []
        for n in self.nodes():
            syns = syns + self.synonyms(n)
        return syns
    
    def label(self, nid, id_if_null=False):
        """
        Fetches label for a node

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        id_if_null : bool
            If True and node has no label return id as label

        Return
        ------
        str
        """
        g = self.get_graph()
        if nid in g:
            n = g.node[nid]
            if 'label' in n:
                return n['label']
            else:
                if id_if_null:
                    return nid
                else:
                    return None
        else:
            if id_if_null:
                return nid
            else:
                return None

    def xrefs(self, nid, bidirectional=False):
        """
        Fetches xrefs for a node

        Arguments
        ---------
        nid : str
            Node identifier for entity to be queried
        bidirection : bool
            If True, include nodes xreffed to nid

        Return
        ------
        list[str]
        """
        if self.xref_graph is not None:
            xg = self.xref_graph
            if nid not in xg:
                return []
            if bidirectional:
                return xg.neighbors(nid)
            else:
                return [x for x in xg.neighbors(nid) if xg[nid][x]['source'] == nid]
                
        return []

    
    def resolve_names(self, names, synonyms=False, **args):
        """
        returns a list of identifiers based on an input list of labels and identifiers.

        Arguments
        ---------
        names: list
           search terms. '%' treated as wildcard
        synonyms: bool
           if true, search on synonyms in addition to labels
        is_regex : bool
           if true, treats each name as a regular expression
        is_partial_match : bool
           if true, treats each name as a regular expression .*name.*
        """
        g = self.get_graph()
        r_ids = []
        for n in names:
            if len(n.split(":")) ==2:
                r_ids.append(n)
            else:
                matches = set([nid for nid in g.nodes() if self._is_match(self.label(nid), n, **args)])
                if synonyms:
                    for nid in g.nodes():
                        for s in self.synonyms(nid):
                            if self._is_match(s.val, n, **args):
                                matches.add(nid) 
                r_ids += list(matches)
        return r_ids

    def _is_match(self, label, term, is_partial_match=False, is_regex=False, **args):
        if term == '%':
            # always match if client passes '%'
            return True
        if label is None:
            label = ''
        if term.find('%') > -1:
            term = term.replace('%','.*')
            is_regex = True
        if is_regex:
            return re.search(term, label) is not None
        if is_partial_match:
            return label.find(term) > -1
        else:
            return label == term
    
    def search(self, searchterm, **args):
        """
        Simple search. Returns list of IDs.

        Arguments
        ---------
        searchterm: list
           search term. '%' treated as wildcard
        synonyms: bool
           if true, search on synonyms in addition to labels
        is_regex : bool
           if true, treats each name as a regular expression
        is_partial_match : bool
           if true, treats each name as a regular expression .*name.*

        Return
        ------
        list
           match node IDs
        """
        return self.resolve_names([searchterm], **args)

class LogicalDefinition():
    """
    A simple OWL logical definition conforming to the pattern:

    ::

        class_id = (genus_id_1 AND ... genus_id_n) AND (P_1 some FILLER_1) AND ... (P_m some FILLER_m)

    See `obographs <https://github.com/geneontology/obographs>`_ docs for more details

    """
    def __init__(self, class_id, genus_ids, restrictions):
        """
        Arguments
        ---------
        class_id : string
            the class that is being defined
        genus_ids : list
            a list of named classes (typically length 1)
        restrictions : list
            a list of (PROPERTY_ID, FILLER_CLASS_ID) tuples
        """
        self.class_id = class_id
        self.genus_ids = genus_ids
        self.restrictions = restrictions
        
    def __str__(self):
        return "{} =def {} AND {}".format(self.class_id, self.genus_ids, self.restrictions)
    def __repr__(self):
        return self.__str__()

class Synonym():
    """
    Represents a synonym using the OBO model
    """

    predmap = dict(
        hasExactSynonym='exact',
        hasBroadSynonym='broad',
        hasNarrowSynonym='narrow',
        hasRelatedSynonym='related')

    def __init__(self, class_id, val=None, pred=None, lextype=None, xrefs=None, ontology=None):
        """
        Arguments:

         - class_id : string

             the class that is being defined

         - value : string

             the synonym itself

         - pred: string

             oboInOwl predicate used to model scope. One of: has{Exact,Narrow,Related,Broad}Synonym - may also be 'label'

         - lextype: string

             From an open ended set of types

         - xrefs: list

             Provenance or cross-references to same usage

        """
        pred = pred.replace("http://www.geneontology.org/formats/oboInOwl#","")
        self.class_id = class_id
        self.val = val
        self.pred = pred
        self.lextype = lextype
        self.xrefs = xrefs
        self.ontology = ontology
        
    def __str__(self):
        return '{} "{}" {} {} {}'.format(self.class_id, self.val, self.pred, self.lextype, self.xrefs)
    def __repr__(self):
        return self.__str__()

    def scope(self):
        return self.predmap[self.pred].upper()
    
    def __cmp__(self, other):
        (x,y) = (str(self),str(other))
        if x > y:
            return 1
        if x < y:
            return -1
        return 0



"""
Parsers for GAF and various Association TSVs.

All parser objects instantiate a subclass of the abstract `AssocParser` object

"""
import re
import requests
import tempfile
from contextlib import closing
import subprocess
import logging

TAXON = 'TAXON'
ENTITY = 'ENTITY'
ANNOTATION = 'ANNOTATION'

class AssocParserConfig():
    """
    Configuration for an association parser
    """
    def __init__(self,
                 remove_double_prefixes=False,
                 entity_map=None,
                 valid_taxa=None,
                 class_idspaces=None):
        self.remove_double_prefixes=remove_double_prefixes
        self.entity_map=entity_map
        self.valid_taxa=valid_taxa
        self.class_idspaces=class_idspaces

class Report():
    """
    A report object that is generated as a result of a parse
    """

    # Levels
    FATAL = 'FATAL'
    ERROR = 'ERROR'
    WARNING = 'WARNING'

    # Warnings: TODO link to gorules
    INVALID_ID = "Invalid identifier"
    INVALID_IDSPACE = "Invalid identifier prefix"
    INVALID_TAXON = "Invalid taxon"
    UNMAPPED_ID = "Unmapped identifier"

    """
    3 warning levels
    """
    LEVELS = [FATAL, ERROR, WARNING]

    def __init__(self):
        self.messages = []
        self.n_lines = 0
        self.n_assocs = 0
        self.skipped = []
        self.subjects = set()
        self.objects = set()
        self.taxa = set()
        self.references = set()

    def error(self, line, type, obj, msg=""):
        self.message(self.ERROR, line, type, obj, msg)
    def warning(self, line, type, obj, msg=""):
        self.message(self.WARNING, line, type, obj, msg)
    def message(self, level, line, type, obj, msg=""):
        self.messages.append({'level':level,
                              'line':line,
                              'type':type,
                              'message':msg,
                              'obj':obj})

    def to_report_json(self):
        """
        Generate a summary in json format
        """

        N = 10
        report = dict(
            summary = dict(association_count = self.n_assocs,
                           line_count = self.n_lines,
                           skipped_line_count = len(self.skipped)),
            aggregate_statistics = dict(subject_count=len(self.subjects),
                                        object_count=len(self.objects),
                                        taxon_count=len(self.taxa),
                                        reference_count=len(self.references),
                                        taxon_sample=list(self.taxa)[0:N],
                                        subject_sample=list(self.subjects)[0:N],
                                        object_sample=list(self.objects)[0:N])
        )

        # grouped messages
        gm = {}
        for level in self.LEVELS:
            gm[level] = []
        for m in self.messages:
            level = m['level']
            gm[level].append(m)

        mgroups = []
        for level in self.LEVELS:
            msgs = gm[level]
            mgroup = dict(level=level,
                          count=len(msgs),
                          messages=msgs)
            mgroups.append(mgroup)
        report['groups'] = mgroups
        return report

    def to_markdown(self):
        """
        Generate a summary in markdown format
        """
        json = self.to_report_json()
        summary = json['summary']

        s = ""
        s += "\n## SUMMARY\n\n"

        s += " * Associations: {}\n" . format(summary['association_count'])
        s += " * Lines in file (incl headers): {}\n" . format(summary['line_count'])
        s += " * Lines skipped: {}\n" . format(summary['skipped_line_count'])

        stats = json['aggregate_statistics']
        s += "\n## STATISTICS\n\n"
        for k,v in stats.items():
            s += " * {}: {}\n" . format(k,v)

        s += "\n## MESSAGES\n\n"
        for g in json['groups']:
            s += " * {}: {}\n".format(g['level'], g['count'])
        s += "\n\n"
        for g in json['groups']:
            level = g['level']
            msgs = g['messages']
            if len(msgs) > 0:
                s += "### {}\n\n".format(level)
                for m in msgs:
                    s += " * {}: {} '{}' `{}`\n".format(m['type'],m['obj'],m['message'],m['line'])
        return s

# TODO avoid using names that are builtin python: file, id

class AssocParser():
    """
    Abstract superclass of all association parser classes
    """

    def parse(self, file, outfile=None):
        """Parse a line-oriented association file into a list of association dict objects

        Note the returned list is of dict objects. TODO: These will
        later be specified using marshmallow and it should be possible
        to generate objects
        
        Arguments
        ---------
        file : file or string
            The file is parsed into association objects. Can be a http URL, filename or `file-like-object`, for input assoc file
        outfile : file
            Optional output file in which processed lines are written. This a file or `file-like-object`

        Return
        ------
        list
            Associations generated from the file
        """
        file = self._ensure_file(file)
        assocs = []
        skipped = []
        n_lines = 0
        for line in file:
            n_lines += 1
            if line.startswith("!"):
                if outfile is not None:
                    outfile.write(line)
                continue
            line = line.strip("\n")
            # Let's rename line2 to something more meaningful
            line2, new_assocs  = self.parse_line(line)
            if new_assocs is None or new_assocs == []:
                logging.warn("SKIPPING: {}".format(new_assocs))
                skipped.append(line)
            else:
                for a in new_assocs:
                    rpt = self.report
                    rpt.subjects.add(a['subject']['id'])
                    rpt.objects.add(a['object']['id'])
                    rpt.references.update(a['evidence']['has_supporting_reference'])
                    if 'taxon' in a['subject']:
                        rpt.taxa.add(a['subject']['taxon']['id'])
                assocs += new_assocs
                if outfile is not None:
                    outfile.write(line2 + "\n")

        self.report.skipped += skipped
        self.report.n_lines += n_lines
        self.report.n_assocs += len(assocs)
        logging.info("Parsed {} assocs from {} lines. Skipped: {}".
                     format(len(assocs),
                            n_lines,
                            len(skipped)))
        file.close()
        return assocs

    def map_to_subset(self, file, outfile=None, ontology=None, subset=None, class_map=None, relations=None):
        """
        Map a file to a subset, writing out results

        You can pass either a subset name (e.g. goslim_generic) or a dictionary with ready-made mappings

        Arguments
        ---------
        file: file
            Name or file object for input assoc file
        outfile: file
            Name or file object for output (mapped) assoc file; writes to stdout if not set
        subset: str
            Optional name of subset to map to, e.g. goslim_generic
        class_map: dict
            Mapping between asserted class ids and ids to map to. Many to many
        ontology: `Ontology`
            Ontology to extract subset from

        """
        if subset is not None:
            logging.info("Creating mapping for subset: {}".format(subset))
            class_map = ontology.create_slim_mapping(subset=subset, relations=relations)

        if class_map is None:
            raise ValueError("Neither class_map not subset is set")
        col = self.ANNOTATION_CLASS_COLUMN
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            logging.info("LINE: {} VALS: {}".format(line, vals))
            if len(vals) < col:
                raise ValueError("Line: {} has too few cols, expect class id in col {}".format(line, col))
            cid = vals[col]
            if cid not in class_map or len(class_map[cid]) == 0:
                self.report.error(line, Report.UNMAPPED_ID, cid)
                continue
            else:
                for mcid in class_map[cid]:
                    vals[col] = mcid
                    line = "\t".join(vals)
                    if outfile is not None:
                        outfile.write(line)
                    else:
                        print(line)
                

    def skim(self, file):
        """
        Lightweight parse of a file into tuples.
        
        Note this discards metadata such as evidence.

        Return a list of tuples (subject_id, subject_label, object_id)
        """
        raise NotImplementedError("AssocParser.skim not implemented")
    
    def parse_line(self, line):
        raise NotImplementedError("AssocParser.parse_line not implemented")

    # split an ID/CURIE into prefix and local parts
    # (not currently used)
    def _parse_id(self, id):
        toks = id.split(":")
        if len(toks) == 2:
            return (toks[0],toks[1])
        else:
            return (toks[0],toks[1:].join(":"))

    # split an ID/CURIE into prefix and local parts
    def _get_id_prefix(self, id):
        toks = id.split(":")
        return toks[0]

    def _validate_taxon(self, taxon, line):
        if self.config.valid_taxa is None:
            return True
        else:
            if taxon in self.config.valid_taxa:
                return True
            else:
                self.report.error(line, Report.INVALID_TAXON, taxon)
                return False

    def _validate_id(self, id, line, context=None):
        if " " in id:
            self.report.error(line, Report.INVALID_ID, id)
            return False
        if id.find("|") > -1:
            # non-fatal
            self.report.error('', Report.INVALID_ID, id)
        idspace = self._get_id_prefix(id)
        if context == ANNOTATION and self.config.class_idspaces is not None:
            if idspace not in self.config.class_idspaces:
                self.report.error(line, Report.INVALID_IDSPACE, id, "allowed: {}".format(self.config.class_idspaces))
                return False
        return True

    def _split_pipe(self, v):
        if v == "":
            return []
        ids = v.split("|")
        ids = [id for id in ids if self._validate_id(id, '')]
        return ids

    def _pair_to_id(self, db, localid):
        if self.config.remove_double_prefixes:
            ## Switch MGI:MGI:n to MGI:n
            if localid.startswith(db+":"):
                localid = localid.replace(db+":","")
        return db + ":" + localid

    def _taxon_id(self,id):
         id = id.replace('taxon','NCBITaxon')
         self._validate_id(id,'',TAXON)
         return id

    def _ensure_file(self, file):
        if isinstance(file,str):
            # TODO Let's fix this if/elseif chain.
            if file.startswith("ftp"):
                f = tempfile.NamedTemporaryFile()
                fn = f.name
                cmd = ['wget',file,'-O',fn]
                subprocess.run(cmd, check=True)
                return open(fn,"r")
            elif file.startswith("http"):
                url = file
                with closing(requests.get(url, stream=False)) as resp:
                    logging.info("URL: {} STATUS: {} ".format(url, resp.status_code))
                    ok = resp.status_code == 200
                    if ok:
                        return io.StringIO(resp.text)
                    else:
                        return None
            else:
                 return open(file, "r")
        else:
            return file


    def _parse_class_expression(self, x):
        ## E.g. exists_during(GO:0000753)
        ## Atomic class expressions only
        [(p,v)] = re.findall('(.*)\((.*)\)',x)
        return {
            'property':p,
            'filler':v
        }

# TODO consider making an Association its own class too to give it a little more
# TODO Semantic value?

# TODO consider making an ID class?


class GpadParser(AssocParser):
    """
    Parser for GO GPAD Format

    https://github.com/geneontology/go-annotation/blob/master/specs/gpad-gpi-1_2.md
    """

    ANNOTATION_CLASS_COLUMN=3
    
    def __init__(self,config=AssocParserConfig()):
        """
        Arguments:
        ---------

        config : a AssocParserConfig object
        """
        self.config = config
        self.report = Report()

    def skim(self, file):
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            if len(vals) != 12:
                logging.error("Unexpected number of columns: {}. GPAD should have 12.".format(vals))
            rel = vals[2]
            # TODO: not
            id = self._pair_to_id(vals[0], vals[1])
            if not self._validate_id(id, line, ENTITY):
                continue
            t = vals[3]
            tuples.append( (id,None,t) )
        return tuples

    def parse_line(self, line):            
        """Parses a single line of a GPAD.

        Return a tuple `(processed_line, associations)`. Typically
        there will be a single association, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GPAD file

        """
        vals = line.split("\t")
        [db,
         db_object_id,
         relation,
         goid,
         reference,
         evidence,
         withfrom,
         interacting_taxon_id, # TODO
         date,
         assigned_by,
         annotation_xp,
         annotation_properties] = vals

        id = self._pair_to_id(db, db_object_id)
        if not self._validate_id(id, line, ENTITY):
            return line, []

        if not self._validate_id(goid, line, ANNOTATION):
            return line, []

        assocs = []
        xp_ors = annotation_xp.split("|")
        for xp_or in xp_ors:
            xp_ands = xp_or.split(",")
            extns = []
            for xp_and in xp_ands:
                if xp_and != "":
                    extns.append(self._parse_class_expression(xp_and))
            assoc = {
                'source_line': line,
                'subject': {
                    'id':id
                },
                'object': {
                    'id':goid,
                    'extensions': extns
                },
                'relation': {
                    'id': relation
                },
                'evidence': {
                    'type': evidence,
                    'with_support_from': self._split_pipe(withfrom),
                    'has_supporting_reference': self._split_pipe(reference)
                },
                'provided_by': assigned_by,
                'date': date,

            }
            assocs.append(assoc)
        return line, assocs

class GafParser(AssocParser):
    """
    Parser for GO GAF format
    """

    ANNOTATION_CLASS_COLUMN=4

    def __init__(self,config=AssocParserConfig()):
        """
        Arguments:
        ---------

        config : a AssocParserConfig object
        """
        self.config = config
        self.report = Report()

    def skim(self, file):
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            if len(vals) < 15:
                logging.error("Unexpected number of vals: {}. GAFv1 has 15, GAFv2 has 17.".format(vals))

            if vals[3] != "":
                continue
            id = self._pair_to_id(vals[0], vals[1])
            if not self._validate_id(id, line, ENTITY):
                continue
            n = vals[2]
            t = vals[4]
            tuples.append( (id,n,t) )
        return tuples


    def parse_line(self, line):
        """
        Parses a single line of a GAF

        Return a tuple `(processed_line, associations)`. Typically
        there will be a single association, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GPAD file

        """
        config = self.config

        vals = line.split("\t")
        # GAF v1 is defined as 15 cols, GAF v2 as 17.
        # We treat everything as GAF2 by adding two blank columns.
        # TODO: check header metadata to see if columns corresponds to declared dataformat version
        if len(vals) == 15:
            vals += ["",""]
        [db,
         db_object_id,
         db_object_symbol,
         qualifier,
         goid,
         reference,
         evidence,
         withfrom,
         aspect,
         db_object_name,
         db_object_synonym,
         db_object_type,
         taxon,
         date,
         assigned_by,
         annotation_xp,
         gene_product_isoform] = vals

        ## --
        ## db + db_object_id. CARD=1
        ## --
        id = self._pair_to_id(db, db_object_id)
        if not self._validate_id(id, line, ENTITY):
            return line, []

        if not self._validate_id(goid, line, ANNOTATION):
            return line, []

        # Example use case: mapping from UniProtKB to MOD ID
        if config.entity_map is not None:
            id = self.map_id(id, config.entity_map)
            toks = id.split(":")
            db = toks[0]
            db_object_id = toks[1:]
            vals[1] = db_object_id

        ## --
        ## end of line re-processing
        ## --
        # regenerate line post-mapping
        line = "\t".join(vals)

        ## --
        ## taxon CARD={1,2}
        ## --
        ## if a second value is specified, this is the interacting taxon
        taxa = [self._taxon_id(x) for x in taxon.split("|")]
        taxon = taxa[0]
        in_taxa = taxa[1:]
        self._validate_taxon(taxon, line)

        ## --
        ## db_object_synonym CARD=0..*
        ## --
        synonyms = db_object_synonym.split("|")
        if db_object_synonym == "":
            synonyms = []

        ## --
        ## process associations
        ## --
        ## note that any disjunctions in the annotation extension
        ## will result in the generation of multiple associations
        assocs = []
        xp_ors = annotation_xp.split("|")
        for xp_or in xp_ors:

            # gather conjunctive expressions in extensions field
            xp_ands = xp_or.split(",")
            extns = []
            for xp_and in xp_ands:
                if xp_and != "":
                    extns.append(self._parse_class_expression(xp_and))

            ## --
            ## qualifier
            ## --
            ## we generate both qualifier and relation field
            relation = None
            qualifiers = qualifier.split("|")
            if qualifier == '':
                qualifiers = []
            negated =  'NOT' in qualifiers
            other_qualifiers = [q for q in qualifiers if q != 'NOT']

            ## In GAFs, relation is overloaded into qualifier.
            ## If no explicit non-NOT qualifier is specified, use
            ## a default based on GPI spec
            if len(other_qualifiers) > 0:
                relation = other_qualifiers[0]
            else:
                if aspect == 'C':
                    relation = 'part_of'
                elif aspect == 'P':
                    relation = 'involved_in'
                elif aspect == 'F':
                    relation = 'enables'
                else:
                    relation = None

            ## --
            ## goid
            ## --
            # TODO We shouldn't overload buildin keywords/functions
            object = {'id':goid,
                      'taxon': taxon}
            
            # construct subject dict
            subject = {
                'id':id,
                'label':db_object_symbol,
                'type': db_object_type,
                'fullname': db_object_name,
                'synonyms': synonyms,
                'taxon': {
                    'id': taxon
                }
            }

            ## --
            ## gene_product_isoform
            ## --
            ## This is mapped to a more generic concept of subject_extensions
            subject_extns = []
            if gene_product_isoform is not None and gene_product_isoform != '':
                subject_extns.append({'property':'isoform', 'filler':gene_product_isoform})

            ## --
            ## evidence
            ## reference
            ## withfrom
            ## --
            evidence = {
                'type': evidence,
                'has_supporting_reference': self._split_pipe(reference)
            }
            evidence['with_support_from'] = self._split_pipe(withfrom)

            ## Construct main return dict
            assoc = {
                'source_line': line,
                'subject': subject,
                'object': object,
                'negated': negated,
                'qualifiers': qualifiers,
                'relation': {
                    'id': relation
                },
                'evidence': evidence,
                'provided_by': assigned_by,
                'date': date,

            }
            if len(subject_extns) > 0:
                assoc['subject_extensions'] = subject_extns
            if len(extns) > 0:
                assoc['object_extensions'] = extns

            assocs.append(assoc)
        return line, assocs

class HpoaParser(GafParser):
    """
    Parser for HPOA format

    http://human-phenotype-ontology.github.io/documentation.html#annot

    Note that there are similarities with Gaf format, so we inherit from GafParser, and override
    """

    def __init__(self,config=AssocParserConfig()):
        """
        Arguments:
        ---------

        config : a AssocParserConfig object
        """
        self.config = config
        self.report = Report()

    def parse_line(self, line):
        """
        Parses a single line of a HPOA file

        Return a tuple `(processed_line, associations)`. Typically
        there will be a single association, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GPAD file

        """
        config = self.config

        # http://human-phenotype-ontology.github.io/documentation.html#annot
        vals = line.split("\t")
        [db,
         db_object_id,
         db_object_symbol,
         qualifier,
         hpoid,
         reference,
         evidence,
         onset,
         frequency,
         withfrom,
         aspect,
         db_object_synonym,
         date,
         assigned_by] = vals

        # hardcode this, as HPOA is currently human-only
        taxon = 'NCBITaxon:9606'

        # hardcode this, as HPOA is currently disease-only
        db_object_type = 'disease'

        ## --
        ## db + db_object_id. CARD=1
        ## --
        id = self._pair_to_id(db, db_object_id)
        if not self._validate_id(id, line, ENTITY):
            return line, []

        if not self._validate_id(hpoid, line, ANNOTATION):
            return line, []


        # Example use case: mapping from OMIM to Orphanet
        if config.entity_map is not None:
            id = self.map_id(id, config.entity_map)
            toks = id.split(":")
            db = toks[0]
            db_object_id = toks[1:]
            vals[1] = db_object_id

        ## --
        ## end of line re-processing
        ## --
        # regenerate line post-mapping
        line = "\t".join(vals)

        ## --
        ## db_object_synonym CARD=0..*
        ## --
        synonyms = db_object_synonym.split("|")
        if db_object_synonym == "":
            synonyms = []


        ## --
        ## qualifier
        ## --
        ## we generate both qualifier and relation field
        relation = None
        qualifiers = qualifier.split("|")
        if qualifier == '':
            qualifiers = []
        negated =  'NOT' in qualifiers
        other_qualifiers = [q for q in qualifiers if q != 'NOT']

        ## CURRENTLY NOT USED
        if len(other_qualifiers) > 0:
            relation = other_qualifiers[0]
        else:
            if aspect == 'O':
                relation = 'has_phenotype'
            elif aspect == 'I':
                relation = 'has_inheritance'
            elif aspect == 'M':
                relation = 'mortality'
            elif aspect == 'C':
                relation = 'has_onset'
            else:
                relation = None

        ## --
        ## hpoid
        ## --
        object = {'id':hpoid,
                  'taxon': taxon}

        # construct subject dict
        subject = {
            'id':id,
            'label':db_object_symbol,
            'type': db_object_type,
            'synonyms': synonyms,
            'taxon': {
                'id': taxon
            }
        }

        ## --
        ## evidence
        ## reference
        ## withfrom
        ## --
        evidence = {
            'type': evidence,
            'has_supporting_reference': self._split_pipe(reference)
        }
        evidence['with_support_from'] = self._split_pipe(withfrom)

        ## Construct main return dict
        assoc = {
            'source_line': line,
            'subject': subject,
            'object': object,
            'negated': negated,
            'qualifiers': qualifiers,
            'relation': {
                'id': relation
            },
            'evidence': evidence,
            'provided_by': assigned_by,
            'date': date,

        }

        return line, [assoc]



from ontobio.ontol_factory import OntologyFactory
from ontobio.lexmap import LexicalMapEngine
import logging

HAS_PART = 'BFO:0000051'
PART_OF = 'BFO:0000050'

QUALITY = 'PATO:0000001'
PLOIDY = 'PATO:0001374'
EUPLOID = 'PATO:0001393'
SHAPE = 'PATO:0000052'
Y_SHAPED = 'PATO:0001201'
PENTAPLOID = 'PATO:0001383'
SWOLLEN = 'PATO:0001851'
DILATED = 'PATO:0001571'
INCREASED_SIZE = 'PATO:0000586'
PROTRUDING = 'PATO:0001598'
MORPHOLOGY = 'PATO:0000051'
ABSENT = 'PATO:0000462'

def test_align():
    """
    Text lexical mapping
    """
    factory = OntologyFactory()
    print("Creating ont")
    ont1 = factory.create('ma')
    ont2 = factory.create('zfa')
    lexmap = LexicalMapEngine()

    lexmap.index_ontology(ont1)
    lexmap.index_ontology(ont2)

    print(lexmap.lmap)
    print(ont1.all_synonyms())
    print(ont2.all_synonyms())
    g = lexmap.get_xref_graph()
    for x,y,d in g.edges_iter(data=True):
        print("{}<->{} :: {}".format(x,y,d))
    for x in g.nodes():
        print("{} --> {}".format(x,lexmap.grouped_mappings(x)))



from enum import Enum


class HomologyTypes(Enum):
    Ortholog = 'RO:HOM0000017'
    LeastDivergedOrtholog = 'RO:HOM0000020'
    Homolog = 'RO:HOM0000007'
    Paralog = 'RO:HOM0000011'
    InParalog = 'RO:HOM0000023'
    OutParalog = 'RO:HOM0000024'
    Ohnolog = 'RO:HOM0000022'
    Xenolog = 'RO:HOM0000018'


from ontobio.ontol_factory import OntologyFactory
import logging
from ontobio.io.ontol_renderers import GraphRenderer

PART_OF = 'BFO:0000050'

QUALITY = 'PATO:0000001'
PLOIDY = 'PATO:0001374'
EUPLOID = 'PATO:0001393'
SHAPE = 'PATO:0000052'
Y_SHAPED = 'PATO:0001201'

def test_local_json_parse():
    """
    Load ontology from JSON
    """
    factory = OntologyFactory()
    print("Creating ont")
    #tbox_ontology = factory.create('go')
    # TODO: create test module for this example
    tbox_ontology = factory.create('tests/resources/go-truncated-pombase.json')
    ont = factory.create('tests/resources/gocam-example.ttl', tbox_ontology=tbox_ontology)
    g = ont.get_graph()
    nodes = ont.search('%')
    print("NODES: {}".format(nodes))
    w = GraphRenderer.create(None)
    w.write_subgraph(ont, nodes)
    i = 'http://model.geneontology.org/0000000300000001/0000000300000007'
    ni = g[i]
    print(str(ni))
    ['GO:0060070'] == ni['types']
    nbrs = ont.neighbors(i)
    print("NEIGHBORS: {}".format(nbrs))
    subont = tbox_ontology.subontology(nodes, minimal=False)
    w = GraphRenderer.create('obo')
    print(w.render(subont))





from ontobio.config import Config, get_config
import logging

def test_cfg():
    logging.info("TEST")
    cfg = get_config()
    logging.info("GOT: {}".format(cfg))
    print(str(cfg))
    assert True
    cfg = get_config()
    print(str(cfg.solr_assocs_url))
    print(str(cfg.ontologies))



from ontobio.golr.golr_associations import top_species
 
def test_species_facet():
    tups = top_species(object_category='function',
                       #url='http://localhost:8983/solr/golr'
    )
    for k,v in tups:
        print("  {:5d}: {}".format(v,k))
   
    



"""
Factory class for generating association sets based on a variety of handle types.

Currently only supports golr query
"""

import networkx as nx
import logging
import os
import subprocess
import hashlib
from cachier import cachier
import datetime
from ontobio.golr.golr_associations import bulk_fetch
from ontobio.assocmodel import AssociationSet, AssociationSetMetadata
import ontobio.io.gafparser as px
from ontobio.io.gafparser import GafParser

SHELF_LIFE = datetime.timedelta(days=3)

class AssociationSetFactory():
    """
    Factory for creating AssociationSets

    Currently support for golr (GO and Monarch) is provided but other stores possible
    """
    def __init__(self):
        """
        initializes based on an ontology name
        """

    def create(self, ontology=None,subject_category=None,object_category=None,evidence=None,taxon=None,relation=None, file=None, fmt=None):
        """
        creates an AssociationSet

        Currently, this uses an eager binding to a `ontobio.golr` instance. All compact associations for the particular combination
        of parameters are fetched.

        Arguments
        ---------

        ontology:         an `Ontology` object
        subject_category: string representing category of subjects (e.g. gene, disease, variant)
        object_category: string representing category of objects (e.g. function, phenotype, disease)
        taxon:           string holding NCBITaxon:nnnn ID
        
        """
        meta = AssociationSetMetadata(subject_category=subject_category,
                                      object_category=object_category,
                                      taxon=taxon)
        
        if file is not None:
            return self.create_from_file(file=file,
                                         fmt=fmt,
                                         ontology=ontology,
                                         meta=meta)

        logging.info("Fetching assocs from store")
        assocs = bulk_fetch_cached(subject_category=subject_category,
                                   object_category=object_category,
                                   evidence=evidence,
                                   taxon=taxon)

        logging.info("Creating map for {} subjects".format(len(assocs)))

        
        
        amap = {}
        subject_label_map = {}
        for a in assocs:
            rel = a['relation']
            subj = a['subject']
            subject_label_map[subj] = a['subject_label']
            amap[subj] = a['objects']

            
        aset = AssociationSet(ontology=ontology,
                              meta=meta,
                              subject_label_map=subject_label_map,
                              association_map=amap)
        return aset

    def create_from_tuples(self, tuples, **args):
        """
        Creates from a list of (subj,subj_name,obj) tuples
        """
        amap = {}
        subject_label_map = {}
        for a in tuples:
            subj = a[0]
            subject_label_map[subj] = a[1]
            if subj not in amap:
                amap[subj] = []
            amap[subj].append(a[2])
        
        aset = AssociationSet(subject_label_map=subject_label_map, association_map=amap, **args)
        return aset
    
    def create_from_file(self, file=None, fmt='gaf', **args):
        """
        Creates from a file.

        Arguments
        ---------
        file : str or file
            input file or filename
        format : str
            name of format e.g. gaf
        
        """
        if isinstance(file,str):
            file = open(file,"r")

        p = None
        if fmt == 'gaf':
            p = px.GafParser()
        elif fmt == 'gpad':
            p = px.GpadParser()
        elif fmt == 'hpoa':
            p = px.HpoaParser()
        else:
            logging.error("Format not recognized: {}".format(fmt))
        logging.info("Parsing {} with {}/{}".format(file, fmt, p))
        results = p.skim(file)
        return self.create_from_tuples(results, **args)
    
    def create_from_gaf(self, file, **args):
        """
        Creates from a GAF file
        """
        p = GafParser()
        results = p.skim(file)
        return self.create_from_tuples(results, **args)


    def create_from_phenopacket(self, file):
        """
        Creates from a phenopacket file
        """
        pass

    def create_from_simple_json(self, file):
        """
        Creates from a simple json rendering
        """
        pass


@cachier(stale_after=SHELF_LIFE)
def bulk_fetch_cached(**args):
        logging.info("Fetching assocs from store (will be cached)")
        return bulk_fetch(**args)
    



#!/usr/bin/env python

"""
Command line wrapper to obographs assocmodel library

Type:

    ogr-assoc -h

For instructions

Examples:

```
ogr-assoc -v -r go -T NCBITaxon:9606 -C gene function enrichment -q GO:1903010 

ogr-assoc -v -r go -T NCBITaxon:10090 -C gene function dendrogram GO:0003700 GO:0005215 GO:0005634 GO:0005737 GO:0005739 GO:0005694 GO:0005730  GO:0000228 GO:0000262 

ogr-assoc -v -r go -T NCBITaxon:10090 -C gene function simmatrix MGI:1890081 MGI:97487 MGI:106593 MGI:97250 MGI:2151057 MGI:1347473

ogr-assoc -C gene function -T pombe -r go -f tests/resources/truncated-pombase.gaf query -q GO:0005622

# requires files from ftp://ftp.rgd.mcw.edu/pub/ontology/annotated_rgd_objects_by_ontology
ogr-assoc  -T . -C gene disease -r doid -f homo_genes_do phenolog -R pw -F homo_genes_pw > PAIRS.txt
```

"""

import argparse
import networkx as nx
from networkx.algorithms.dag import ancestors, descendants
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory
from ontobio.io.ontol_renderers import GraphRenderer
from ontobio.slimmer import get_minimal_subgraph
import logging

def main():
    """
    Wrapper for OGR Assocs
    """
    parser = argparse.ArgumentParser(description='Wrapper for obographs assocmodel library'
                                                 """
                                                 By default, ontologies and assocs are cached locally and synced from a remote sparql endpoint
                                                 """,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-r', '--resource', type=str, required=False,
                        help='Name of ontology')
    parser.add_argument('-f', '--assocfile', type=str, required=False,
                        help='Name of input file for associations')
    parser.add_argument('--assocformat', type=str, default='gaf', required=False,
                        help='Format of association file, if passed (default: gaf)')
    parser.add_argument('-o', '--outfile', type=str, required=False,
                        help='Path to output file')
    parser.add_argument('-t', '--to', type=str, required=False,
                        help='Output to (tree, dot, ...)')
    parser.add_argument('-d', '--direction', type=str, default='u', required=False,
                        help='u = up, d = down, ud = up and down')
    parser.add_argument('-e', '--evidence', type=str, required=False,
                        help='ECO class')
    parser.add_argument('-p', '--properties', nargs='*', type=str, required=False,
                        help='Properties')
    parser.add_argument('-s', '--search', type=str, default='', required=False,
                        help='Search type. p=partial, r=regex')
    parser.add_argument('-S', '--slim', type=str, default='', required=False,
                        help='Slim type. m=minimal')
    parser.add_argument('-c', '--container_properties', nargs='*', type=str, required=False,
                        help='Properties to nest in graph')
    parser.add_argument('-C', '--category', nargs=2, type=str, required=True,
                        help='category tuple (SUBJECT OBJECT)')
    parser.add_argument('-T', '--taxon', type=str, required=True,
                        help='Taxon of associations')
    parser.add_argument('-v', '--verbosity', default=0, action='count',
                        help='Increase output verbosity')

    subparsers = parser.add_subparsers(dest='subcommand', help='sub-command help')

    # EXTRACT ONTOLOGY
    parser_n = subparsers.add_parser('subontology', help='Extract sub-ontology')
    parser_n.add_argument('-M', '--minimal', dest='minimal', action='store_true', default=False, help='If set, remove non-MRCA nodes')
    parser_n.set_defaults(function=extract_ontology)
    
    # ENRICHMENT
    parser_n = subparsers.add_parser('enrichment', help='Perform an enrichment test')
    parser_n.add_argument('-q', '--query',type=str, help='query all genes for this class an use as subject')
    parser_n.add_argument('-H', '--hypotheses',nargs='*', help='list of classes to test against')
    parser_n.add_argument('subjects',nargs='*')
    parser_n.set_defaults(function=run_enrichment_test)

    # PHENOLOG
    parser_n = subparsers.add_parser('phenolog', help='Perform multiple enrichment tests')
    parser_n.add_argument('-R', '--resource2',type=str, required=True, help='path to second GAF')
    parser_n.add_argument('-F', '--file2',type=str, required=True, help='handle for second ontology')
    parser_n.set_defaults(function=run_phenolog)
    
    # QUERY
    parser_n = subparsers.add_parser('query', help='Query based on positive and negative terms')
    parser_n.add_argument('-q', '--query',nargs='*', help='positive classes')
    parser_n.add_argument('-N', '--negative',type=str, help='negative classes')
    parser_n.set_defaults(function=run_query)
    
    # QUERY ASSOCIATIONS
    parser_n = subparsers.add_parser('associations', help='Query for association pairs')
    parser_n.add_argument('subjects',nargs='*', help='subject ids')
    parser_n.set_defaults(function=run_query_associations)

    # INTERSECTIONS
    parser_n = subparsers.add_parser('intersections', help='Query intersections')
    parser_n.add_argument('-X', '--xterms',nargs='*', help='x classes')
    parser_n.add_argument('-Y', '--yterms',nargs='*', help='y classes')
    parser_n.add_argument('--useids',type=bool, default=False, help='if true, use IDs not labels on axes')
    parser_n.add_argument('terms',nargs='*', help='all terms (x and y)')
    parser_n.set_defaults(function=plot_intersections)

    # INTERSECTION DENDROGRAM
    parser_n = subparsers.add_parser('dendrogram', help='Plot dendrogram from intersections')
    parser_n.add_argument('-X', '--xterms',nargs='*', help='x classes')
    parser_n.add_argument('-Y', '--yterms',nargs='*', help='y classes')
    parser_n.add_argument('--useids',type=bool, default=False, help='if true, use IDs not labels on axes')
    parser_n.add_argument('terms',nargs='*', help='all terms (x and y)')
    parser_n.set_defaults(function=plot_term_intersection_dendrogram)

    # SIMILARITY MATRIX
    parser_n = subparsers.add_parser('simmatrix', help='Plot dendrogram for similarities between subjects')
    parser_n.add_argument('-X', '--xsubjects',nargs='*', help='x subjects')
    parser_n.add_argument('-Y', '--ysubjects',nargs='*', help='y subjects')
    parser_n.add_argument('--useids',type=bool, default=False, help='if true, use IDs not labels on axes')
    parser_n.add_argument('subjects',nargs='*', help='all terms (x and y)')
    parser_n.set_defaults(function=plot_simmatrix)

    args = parser.parse_args()

    if args.verbosity >= 2:
        logging.basicConfig(level=logging.DEBUG)
    if args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    logging.info("Welcome!")
    
    handle = args.resource

    # Ontology Factory
    ofactory = OntologyFactory()
    logging.info("Creating ont object from: {} {}".format(handle, ofactory))
    ont = ofactory.create(handle)
    logging.info("ont: {}".format(ont))

    evidence = args.evidence
    if evidence is not None and evidence.lower() == 'noiea':
        evidence = "-ECO:0000501"
    
    
    # Association Factory
    afactory = AssociationSetFactory()
    [subject_category, object_category] = args.category
    aset = None
    if args.assocfile is not None:
        aset = afactory.create_from_file(file=args.assocfile,
                                         fmt=args.assocformat,
                                         ontology=ont)
    else:
        # create using GO/Monarch services
        aset = afactory.create(ontology=ont,
                               subject_category=subject_category,
                               object_category=object_category,
                               taxon=args.taxon)
    
    
    func = args.function
    func(ont, aset, args)

def extract_ontology(ont, aset, args):
    ont = aset.subontology(minimal=args.minimal)
    w = GraphRenderer.create('obo')
    w.outfile = args.outfile
    w.write(ont)
    
def run_enrichment_test(ont, aset, args):
    subjects = args.subjects
    if args.query is not None:
        subjects = aset.query([args.query])
    print("SUBJECTS q={} : {}".format(args.query, subjects))
    enr = aset.enrichment_test(subjects=subjects, hypotheses=args.hypotheses, labels=True)
    for r in enr:
        print("{:8.3g} {} {:40s}".format(r['p'],r['c'],str(r['n'])))

def run_phenolog(ont, aset, args):
    ofactory = OntologyFactory()
    ont2 = ofactory.create(args.resource2)

    afactory = AssociationSetFactory()
    aset2 = afactory.create(ontology=ont2,
                            file=args.file2)

    common = set(aset.subjects).intersection(aset2.subjects)
    num_common = len(common)
    logging.info("Genes in common between two KBs: {}/\{} = {}".format(len(aset.subjects), len(aset2.subjects), num_common))
    if num_common < 2:
        logging.error("TOO FEW")
        return None
    for n in aset.ontology.nodes():
        nl = ont.label(n, id_if_null=True)
        genes = aset.query([n])
        num_genes = len(genes)
        if num_genes > 2:
            logging.info("BASE: {} {} num={}".format(n,nl, num_genes))
            enr = aset2.enrichment_test(subjects=genes, background=aset2.subjects, labels=True)
            for r in enr:
                print("{:8.3g} {} {:20s} <-> {} {:20s}".format(r['p'],n,nl,r['c'],str(r['n'])))

        
def run_query(ont, aset, args):
    import plotly.plotly as py
    import plotly.graph_objs as go
    subjects = aset.query(args.query, args.negative)
    for s in subjects:
        print("{} {}".format(s, str(aset.label(s))))
    tups = aset.query_associations(subjects=subjects)
    z, xaxis, yaxis = tuple_to_matrix(tups)
    spacechar = " "
    xaxis = mk_axis(xaxis, aset, args, spacechar=" ")
    yaxis = mk_axis(yaxis, aset, args, spacechar=" ")
    logging.info("PLOTTING: {} x {} = {}".format(xaxis, yaxis, z))
    trace = go.Heatmap(z=z,
                       x=xaxis,
                       y=yaxis)
    data=[trace]
    py.plot(data, filename='labelled-heatmap')

def run_query_associations(ont, aset, args):
    import plotly.plotly as py
    import plotly.graph_objs as go
    tups = aset.query_associations(subjects=args.subjects)
    for (s,c) in tups:
        print("{} {}".format(s, c))
    z, xaxis, yaxis = tuple_to_matrix(tups)
    logging.info("PLOTTING: {} x {} = {}".format(xaxis, yaxis, z))
    trace = go.Heatmap(z=z,
                       x=xaxis,
                       y=yaxis)
    data=[trace]
    py.plot(data, filename='labelled-heatmap')
    #plot_dendrogram(z, xaxis, yaxis)

def tuple_to_matrix(tups):
    import numpy as np
    xset = set()
    yset = set()
    for (x,y) in tups:
        xset.add(x)
        yset.add(y)

    xset = list(xset)
    yset = list(yset)
    xmap = {}
    xi = 0
    for x in xset:
        xmap[x] = xi
        xi = xi+1
        
    ymap = {}
    yi = 0
    for y in yset:
        ymap[y] = yi
        yi = yi+1

    logging.info("Making {} x {}".format(len(xset), len(yset)))
    z = [ [0] * len(xset) for i1 in range(len(yset)) ]
        
    for (x,y) in tups:
        z[ymap[y]][xmap[x]] = 1
    z = np.array(z)
    #z = -z
    return (z, xset, yset)

        
def create_intersection_matrix(ont, aset, args):
    xterms = args.xterms
    yterms = args.yterms
    if args.terms is not None and len(args.terms) > 0:
        xterms = args.terms
    if yterms is None or len(yterms) == 0:
        yterms = xterms
    logging.info("X={} Y={}".format(xterms,yterms))
    ilist = aset.query_intersections(x_terms=xterms, y_terms=yterms)
    z, xaxis, yaxis = aset.intersectionlist_to_matrix(ilist, xterms, yterms)
    xaxis = mk_axis(xaxis, aset, args)
    yaxis = mk_axis(yaxis, aset, args)
    return (z, xaxis, yaxis)

def plot_intersections(ont, aset, args):
    import plotly.plotly as py
    import plotly.graph_objs as go
    (z, xaxis, yaxis) = create_intersection_matrix(ont, aset, args)
    trace = go.Heatmap(z=z,
                       x=xaxis,
                       y=yaxis)
    data=[trace]
    py.plot(data, filename='labelled-heatmap')

def plot_simmatrix(ont, aset, args):
    import numpy as np
    xsubjects = args.xsubjects
    ysubjects = args.ysubjects
    if args.subjects is not None and len(args.subjects) > 0:
        xsubjects = args.subjects
    if ysubjects is None or len(ysubjects) == 0:
        ysubjects = xsubjects
    (z, xaxis, yaxis) = aset.similarity_matrix(xsubjects, ysubjects)
    xaxis = mk_axis(xaxis, aset, args)
    yaxis = mk_axis(yaxis, aset, args)
    z = np.array(z)
    z = -z
    print("Z={}".format(z))
    plot_dendrogram(z, xaxis, yaxis)
    
def plot_term_intersection_dendrogram(ont, aset, args):
    import numpy as np
    # TODO: currently only works for xaxis=yaxis
    (z, xaxis, yaxis) = create_intersection_matrix(ont, aset, args)
    z = np.array(z)
    z = -z
    print("Z={}".format(z))
    plot_dendrogram(z, xaxis, yaxis)
    
def plot_dendrogram(z, xaxis, yaxis):
    import plotly.plotly as py
    import plotly.figure_factory as FF
    import plotly.graph_objs as go
    import numpy as np
    from scipy.cluster.hierarchy import linkage
    from scipy.spatial.distance import pdist, squareform


    # Initialize figure by creating upper dendrogram
    figure = FF.create_dendrogram(z, orientation='bottom', labels=xaxis)
    for i in range(len(figure['data'])):
        figure['data'][i]['yaxis'] = 'y2'

    # Create Side Dendrogram
    # TODO: figure out how to create labels for this axis
    dendro_side = FF.create_dendrogram(z, orientation='right', labels=xaxis)
    for i in range(len(dendro_side['data'])):
        dendro_side['data'][i]['xaxis'] = 'x2'
        
    # Add Side Dendrogram Data to Figure
    figure['data'].extend(dendro_side['data'])

    # Create Heatmap
    dendro_leaves = dendro_side['layout']['yaxis']['ticktext']
    #dendro_leaves = list(map(int, dendro_leaves))
    data_dist = pdist(z)
    heat_data = squareform(data_dist)
    #heat_data = heat_data[dendro_leaves,:]
    #heat_data = heat_data[:,dendro_leaves]
    
    heatmap = go.Data([
        go.Heatmap(
            x = dendro_leaves,
            y = dendro_leaves,
            z = heat_data,
            colorscale = 'YIGnBu'
        )
    ])
    heatmap[0]['x'] = figure['layout']['xaxis']['tickvals']
    heatmap[0]['y'] = dendro_side['layout']['yaxis']['tickvals']

    # Add Heatmap Data to Figure
    figure['data'].extend(go.Data(heatmap))

    # Edit Layout
    figure['layout'].update({'width':800, 'height':800,
                         'showlegend':False, 'hovermode': 'closest',
                         })
    # Edit xaxis
    figure['layout']['xaxis'].update({'domain': [.15, 1],
                                  'mirror': False,
                                  'showgrid': False,
                                  'showline': False,
                                  'zeroline': False,
                                  'ticks':""})
    # Edit xaxis2
    figure['layout'].update({'xaxis2': {'domain': [0, .15],
                                   'mirror': False,
                                   'showgrid': False,
                                   'showline': False,
                                   'zeroline': False,
                                   'showticklabels': False,
                                   'ticks':""}})

    # Edit yaxis
    figure['layout']['yaxis'].update({'domain': [0, .85],
                                  'mirror': False,
                                  'showgrid': False,
                                  'showline': False,
                                  'zeroline': False,
                                  'showticklabels': False,
                                  'ticks': ""})
    # Edit yaxis2
    figure['layout'].update({'yaxis2':{'domain':[.825, .975],
                                   'mirror': False,
                                   'showgrid': False,
                                   'showline': False,
                                   'zeroline': False,
                                   'showticklabels': False,
                                   'ticks':""}})

    py.plot(figure, filename='dendrogram_with_labels')
    
def mk_axis(terms, kb, args, spacechar="<br>"):
    # TODO - more elegant solution to blank node hack
    return [label_or_id(x, kb).replace(" ",spacechar).replace("some variant of ","") for x in terms]

def label_or_id(x, kb):
    label = kb.label(x)
    if label is None:
        return x
    else:
        return label

if __name__ == "__main__":
    main()



from prefixcommons.curie_util import contract_uri, expand_uri, get_prefixes
from ontobio.vocabulary.relations import OboRO, Evidence
from ontobio.vocabulary.upper import UpperLevel
from rdflib import Namespace
from rdflib import BNode
from rdflib import Literal
from rdflib import URIRef
from rdflib.namespace import RDF
from rdflib.namespace import RDFS
from rdflib.namespace import OWL
import rdflib
import logging

ro = OboRO()
evt = Evidence()
upt = UpperLevel()

class RdfGenerator():

    def __init__(self):
        self.graph = rdflib.Graph()
        self.include_subject_info = False

    def genid(self):
        return BNode()

    def blanknode(self):
        return BNode()
        
    def uri(self,id):
        # allow either atoms or objects
        if isinstance(id,dict):
            return self.uri(id['id'])
        logging.info("Expand: {}".format(id))
        return URIRef(expand_uri(id))
    
    def emit(self,s,p,o):
        self.graph.add((s,p,o))
        return (s,p,o)
    
    def emit_type(self,s,t):
        return self.emit(s,RDF.type,t)
    def emit_label(self,s,t):
        return self.emit(s,RDFS.label,o)
    
    def translate_evidence(self, association, stmt):
        """

        ``
        _:1 a Axiom
            subject s
            predicate p
            object o
            evidence [ a ECO ; ...]
        ``
        
        """
        ev = association['evidence']
        ev_id = None
        if 'id' in ev:
            ev_id = self.uri(ev['id'])
        else:
            ev_id = self.genid()

        stmt_id = self.blanknode() ## OWL reification: must be blank
        (s,p,o) = stmt
        self.emit_type(stmt_id, OWL.Axiom)
        
        self.emit(stmt_id, OWL.subject, s)        
        self.emit(stmt_id, OWL.predicate, p)        
        self.emit(stmt_id, OWL.object, o)
        
        self.emit(stmt_id, self.uri(evt.axiom_has_evidence), ev_id)
        self.emit_type(ev_id, self.uri(ev['type']))
        if 'with_support_from' in ev:
            pass # TODO
        for ref in ev['has_supporting_reference']:
            self.emit(ev_id, self.uri(evt.has_supporting_reference), self.uri(ref))
        if 'with_support_from' in ev:
            for ref in ev['with_support_from']:
                self.emit(ev_id, self.uri(evt.evidence_with_support_from), self.uri(ref))
        

class CamGenerator(RdfGenerator):
    """
    Granular instance-based representation (GO-CAM)
    """

    def translate(self, association):
        sub = association['subject']
        obj = association['subject']
        rel = association['relation']
        sub_uri = self.uri(sub)
        obj_uri = self.uri(obj)

        enabler_id = self.genid()
        self.emit_type(enabler_id, sub_uri)
        tgt_id = self.genid()
        self.emit_type(tgt_id, obj_uri)
        
        aspect = association['aspect']
        stmt = None
        if aspect == 'F':
            stmt = self.emit(tgt_id, ro.enabled_by, enabler_id)
        elif aspect == 'P':
            mf_id = self.genid()
            self.emit_type(tgt_id, upt.molecular_function)
            stmt = self.emit(tgt_id, ro.enabled_by, mf_id)
            stmt = self.emit(mf_id, ro.part_of, tgt_id)
        elif aspect == 'C':
            mf_id = self.genid()
            self.emit_type(tgt_id, upt.molecular_function)
            stmt = self.emit(tgt_id, ro.enabled_by, mf_id)
            stmt = self.emit(mf_id, ro.occurs_in, tgt_id)

        if self.include_subject_info:
            pass
            # TODO
        # TODO: extensions
        self.translate_evidence(association, stmt)
        
class SimpleAssocGenerator(RdfGenerator):
    """
    Follows simple OBAN-style model
    """

    def translate(self, association):
        sub = association['subject']
        obj = association['subject']
        rel = association['relation']
        sub_uri = self.uri(sub)
        obj_uri = self.uri(obj)
        rel_uri = self.uri(rel)

        # TODO: extensions
        stmt = self.emit(sub_uri,rel_uri,obj_uri)

        # optionally include info about subject (e.g. gene)
        if self.include_subject_info:
            self.emit_label(sub_uri, sub)
            if 'taxon' in sub:
                taxon = sub['taxon']
                self.emit(sub_uri, ro.in_taxon, self.uri(taxon))
            # TODO syns etc

        self.translate_evidence(association, stmt)
        



import rdflib
from rdflib import Namespace
from rdflib.namespace import RDF
from rdflib.namespace import RDFS
from rdflib.namespace import OWL
from rdflib import URIRef
import networkx

from prefixcommons.curie_util import contract_uri
from ontobio.ontol import LogicalDefinition
from ontobio.ontol import Ontology
import ontobio.ontol

def id(uri):
    curies = contract_uri(uri)
    if len(curies) > 0:
        return curies[0]
    else:
        return uri.toPython()

class RdfMapper():
    def __init__(self, rdfgraph=None, tbox_ontology=None):
        self.rdfgraph = rdfgraph
        self.tbox_ontology = tbox_ontology

    def parse_rdf(self,filename=None, format='ttl'):
        self.rdfgraph = rdflib.Graph()
        self.rdfgraph.parse(filename, format=format)

    def convert(self, filename=None, format='ttl'):
        if filename is not None:
            self.parse_rdf(filename=filename, format=format)
        g = networkx.MultiDiGraph()
        ont = Ontology(graph=g)
        self.add_triples(ont)
        return ont
        
    def add_triples(self, ontol):
        """
        Adds triples to an ontology object.

        Currently assumes gocam/lego-style
        """
        rg = self.rdfgraph
        g = ontol.get_graph()
        typemap = {}
        inds = rg.subjects(RDF.type, OWL.NamedIndividual)
        for s in inds:
            for (s,p,o) in rg.triples((s,None,None)):
                s_id = id(s)
                p_id = id(p)
                g.add_node(s_id)
                if isinstance(o,URIRef):
                    o_id = id(o)
                    if p == RDF.type:
                        if o != OWL.NamedIndividual:
                            if s_id not in typemap:
                                typemap[s_id] = []
                            typemap[s_id].append(o_id)
                    else:
                        g.add_edge(o_id,s_id,pred=p_id)

        # propagate label from type
        for s in typemap.keys():
            g[s]['types'] = typemap[s]
            if self.tbox_ontology is not None:
                if 'label' not in g[s]:
                    g[s]['label'] = ";".join([self.tbox_ontology.label(x) for x in typemap[s] if self.tbox_ontology.label(x) is not None])
            
        



import networkx

LABEL_OR_EXACT = 'label_or_exact'

class LexicalMapEngine():
    """
    generates lexical matches between pairs of ontology classes
    """
    def __init__(self):
        self.lmap = {}
        self.smap = {}

    def index_ontology(self, ont):
        """
        Adds an ontology to the index
        """
        syns = ont.all_synonyms(include_label=True)
        for syn in syns:
            self.index_synonym(syn, ont)
            # TODO: stem

    def index_synonym(self, syn, ont):
        syn.ontology = ont
        lmap = self.lmap
        smap = self.smap
        v = syn.val
        cid = syn.class_id
        if v not in lmap:
            lmap[v] = []
        lmap[v].append(syn)
        if cid not in smap:
            smap[cid] = []
        smap[cid].append(syn)

    def find_equiv_sets(self):
        return self.lmap

    def get_xref_graph(self):
        g = networkx.MultiGraph()
        for (v,syns) in self.lmap.items():
            for s1 in syns:
                for s2 in syns:
                    if s1.class_id < s2.class_id:
                        g.add_edge(s1.class_id, s2.class_id, syns=(s1,s2))
        xg = networkx.MultiGraph()
        for i in g.nodes():
            for j in networkx.neighbors(g,i):
                best = 0
                bestm = None
                for ix,m in g[i][j].items():
                    (s1,s2) = m['syns']
                    score = self.combine_syns(s1,s2)
                    if score > best:
                        best = score
                        bestm = m
                xg.add_edge(i,j,best=best)

        return xg
    
    def grouped_mappings(self,id):
        g = self.get_xref_graph()
        m = {}
        for n in g.neighbors(id):
            [prefix, local] = n.split(':')
            if prefix not in m:
                m[prefix] = []
            m[prefix].append(n)
        return m

    def trim_xref_graph(self,g):
        g = self.get_xref_graph()
        for n in g.nodes():
            g.neighbors()

    def combine_syns(self, s1,s2):
        cpred = self.combine_preds(s1.pred, s1.pred)
        s = self.pred_score(cpred)
        return s
    
    def rollup(p):
        if p == 'label':
            return LABEL_OR_EXACT
        if p == 'hasExactSynonym':
            return LABEL_OR_EXACT
        return p
    
    def combine_preds(self, p1, p2):
        if p1 == p2:
            return p1
        if rollup(p1) == rollup(p2):
            return rollup(p1)
        return p1 + p2

    def pred_score(self,p):
        if p == 'label':
            return 10
        if p == LABEL_OR_EXACT:
            return 9
        if p == 'hasExactSynonym':
            return 9
        return 5
                        
        



import networkx as nx
import logging


def get_minimal_subgraph(g, nodes):
    """
    given a set of nodes, extract a subgraph that excludes non-informative nodes - i.e.
    those that are not MRCAs of pairs of existing nodes.
    
    Note: no property chain reasoning is performed. As a result, edge labels are lost.
    """
    logging.info("Slimming {} to {}".format(g,nodes))
    # maps ancestor nodes to members of the focus node set they subsume
    mm = {}
    subnodes = set()
    for n in nodes:
        subnodes.add(n)
        ancs = nx.ancestors(g, n)
        ancs.add(n)
        for a in ancs:
            subnodes.add(a)
            if a not in mm:
                mm[a] = set()
            mm[a].add(n)

    # merge graph
    egraph = nx.MultiDiGraph()

    # TODO: ensure edge labels are preserved
    for a, aset in mm.items():
        for p in g.predecessors(a):
            logging.info(" cmp {} -> {} // {} {}".format(len(aset),len(mm[p]), a, p))
            if p in mm and len(aset) == len(mm[p]):
                egraph.add_edge(p, a)
                egraph.add_edge(a, p)
                logging.info("will merge {} <-> {} (members identical)".format(p,a))

    nmap = {}
    leafmap = {}
    disposable = set()
    for cliq in nx.strongly_connected_components(egraph):
        leaders = set()
        leafs = set()
        for n in cliq:
            is_src = False
            if n in nodes:
                logging.info("Preserving: {} in {}".format(n,cliq))
                leaders.add(n)
                is_src = True
                
            is_leaf = True
            for p in g.successors(n):
                if p in cliq:
                    is_leaf = False

            if not(is_leaf or is_src):
                disposable.add(n)
                
            if is_leaf:
                logging.info("Clique leaf: {} in {}".format(n,cliq))
                leafs.add(n)

                
        leader = None
        if len(leaders) > 1:
            logging.info("UHOH: {}".format(leaders))
        if len(leaders) > 0:
            leader = list(leaders)[0]
        else:
            leader = list(leafs)[0]
        leafmap[n] = leafs

    subg = g.subgraph(subnodes)
    fg = remove_nodes(subg, disposable)
    return fg

def remove_nodes(g, rmnodes):
    logging.info("Removing {} from {}".format(rmnodes,g))
    newg = nx.MultiDiGraph()
    for n,nd in g.nodes(data=True):
        if n not in rmnodes:
            newg.add_node(n, attr_dict=nd)
            parents = _traverse(g, set([n]), set(rmnodes), set())
            for p in parents:
                newg.add_edge(p,n,attr_dict={'pred':'subClassOf'})
    return newg     

def _traverse(g, nset, rmnodes, acc):
    if len(nset) == 0:
        return acc
    n = nset.pop()
    parents = set(g.predecessors(n))
    acc = acc.union(parents - rmnodes)
    nset = nset.union(parents.intersection(rmnodes))
    return _traverse(g, nset, rmnodes, acc)




#!/usr/bin/env python

"""
Command line wrapper to gafparser.py

Type:

    ogr-parse-assocs -h

For instructions

Examples:

```
```

"""

import argparse
import networkx as nx
from networkx.algorithms.dag import ancestors, descendants
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory
from ontobio.io.gafparser import GafParser
from ontobio.slimmer import get_minimal_subgraph
import logging

def main():
    """
    Wrapper for Assoc Parsing
    """
    parser = argparse.ArgumentParser(description='Wrapper for obographs assocmodel library'
                                                 """
                                                 By default, ontologies and assocs are cached locally and synced from a remote sparql endpoint
                                                 """,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-r', '--resource', type=str, required=False,
                        help='Name of ontology')
    parser.add_argument('-f', '--file', type=str, required=False,
                        help='Name of input file for associations - currently GAF is assumed')
    parser.add_argument('-o', '--outfile', type=str, required=False,
                        help='Path to output file')
    parser.add_argument('-m', '--messagefile', type=str, required=False,
                        help='Path to messages (report) markdown file')
    parser.add_argument('-t', '--to', type=str, required=False,
                        help='Output to (tree, dot, ...)')
    parser.add_argument('-T', '--taxon', nargs='*', required=False,
                        help='valid taxon (NCBITaxon ID) - validate against this')
    parser.add_argument('--subject_prefix', nargs='*', required=False,
                        help='E.g PomBase - validate against this')
    parser.add_argument('--object_prefix', nargs='*', required=False,
                        help='E.g GO - validate against this')
    parser.add_argument('-v', '--verbosity', default=0, action='count',
                        help='Increase output verbosity')

    subparsers = parser.add_subparsers(dest='subcommand', help='sub-command help')
    
    parser_n = subparsers.add_parser('validate', help='Validate associations')
    parser_n.set_defaults(function=validate_assocs)

    parser_n = subparsers.add_parser('filter', help='Filter associations')
    parser_n.set_defaults(function=filter_assocs)

    parser_n = subparsers.add_parser('map2slim', help='Map to a subset/slim')
    parser_n.set_defaults(function=map2slim)
    parser_n.add_argument('-p', '--properties', nargs='*', type=str, required=False,
                          help='Properties')
    parser_n.add_argument('-s', '--subset', type=str, required=True,
                          help='subset (e.g. map2slim)')
    
    args = parser.parse_args()

    if args.verbosity >= 2:
        logging.basicConfig(level=logging.DEBUG)
    if args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    logging.info("Welcome!")
    
    handle = args.resource

    # Ontology Factory
    ofactory = OntologyFactory()
    logging.info("Creating ont object from: {} {}".format(handle, ofactory))
    ont = ofactory.create(handle)
    logging.info("ont: {}".format(ont))

    func = args.function
    p = GafParser()

    # set configuration
    config = p.config
    config.valid_taxa = args.taxon
    config.class_idspaces = args.object_prefix
    config.entity_idspaces = args.subject_prefix
    
    outfh = None
    if args.outfile is not None:
        outfh = open(args.outfile, "w")
    func(ont, args.file, outfh,  p, args)
    if outfh is not None:
        outfh.close()
    if args.messagefile is not None:
        mfh = open(args.messagefile, "w")
        mfh.write(p.report.to_markdown())
        mfh.close()
    else:
        print(p.report.to_markdown())
        
def filter_assocs(ont, file, outfile, p, args):
    assocs = p.parse(open(file, "r"), outfile)

def validate_assocs(ont, file, outfile, p, args):
    assocs = p.parse(open(file, "r"), outfile)

def map2slim(ont, file, outfile, p, args):
    assocs = p.map_to_subset(open(file, "r"),
                             ontology=ont, outfile=outfile, subset=args.subset, relations=args.properties)
    


    


if __name__ == "__main__":
    main()



from ontobio.io.gafparser import GafParser, GpadParser
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory
import tempfile
import logging

POMBASE = "tests/resources/truncated-pombase.gaf"
ONT = "tests/resources/go-truncated-pombase.json"

SUBSET = [
    'GO:0051169', ## nuclear transport
    'GO:0006810', ## transport
    'GO:0005730', ## nucleolus
    'GO:0016604', ## nuclear body
    'GO:0023052'  ## signaling
]
    
def test_map2slim_gaf():
    f = POMBASE
    p = GafParser()
    is_gaf = f == POMBASE
    ont = OntologyFactory().create(ONT)
    relations=['subClassOf', 'BFO:0000050']
    m = ont.create_slim_mapping(subset_nodes=SUBSET, relations=relations)
    print(str(m))
    assert m['GO:0071423'] == ['GO:0006810']
    outfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
    p.map_to_subset(open(f,"r"), class_map=m, outfile=outfile)
    for m in p.report.messages:
        print("MESSAGE: {}".format(m))
    p = GafParser()
    outfile.close()
    logging.info("Reading from: {}".format(outfile.name))
    assocs = p.parse(outfile.name)
    assert len(assocs) > 100
    cls_ids = set()
    for a in assocs:
        cid = a['object']['id']
        assert cid in SUBSET
        cls_ids.add(cid)
        print(str(a))
    print(cls_ids)



"""
Mapping between obograph-JSON format and networkx
"""

import itertools
import re
import json

import networkx
import logging
from prefixcommons.curie_util import contract_uri
from ontobio.ontol import LogicalDefinition

def contract_uri_wrap(uri):
    curies = contract_uri(uri)
    if len(curies) > 0:
        return curies[0]
    else:
        return uri

def add_obograph_digraph(og, digraph, node_type=None, predicates=None, xref_graph=None, logical_definitions=None, parse_meta=True, **args):
    """
    Converts a single obograph to Digraph edges and adds to an existing networkx DiGraph
    """
    logging.info("NODES: {}".format(len(og['nodes'])))

    # if client passes an xref_graph we must parse metadata
    if xref_graph is not None:
        parse_meta = True
        
    for n in og['nodes']:
        is_obsolete =  'is_obsolete' in n and n['is_obsolete'] == 'true'
        if is_obsolete:
            continue
        if node_type is not None and ('type' not in n or n['type'] != node_type):
            continue
        id = contract_uri_wrap(n['id'])
        digraph.add_node(id, attr_dict=n)
        if 'lbl' in n:
            digraph.node[id]['label'] = n['lbl']
        if parse_meta and 'meta' in n:
            meta = n['meta']
            if xref_graph is not None and 'xrefs' in meta:
                for x in meta['xrefs']:
                    xref_graph.add_edge(contract_uri_wrap(x['val']), id, source=id)
    logging.info("EDGES: {}".format(len(og['edges'])))
    for e in og['edges']:
        sub = contract_uri_wrap(e['sub'])
        obj = contract_uri_wrap(e['obj'])
        pred = contract_uri_wrap(e['pred'])
        if pred == 'is_a':
            pred = 'subClassOf'
        if predicates is None or pred in predicates:
            digraph.add_edge(obj, sub, pred=pred)
    if 'equivalentNodesSets' in og:
        nslist = og['equivalentNodesSets']
        logging.info("CLIQUES: {}".format(len(nslist)))
        for ns in nslist:
            equivNodeIds = ns['nodeIds']
            for i in ns['nodeIds']:
                ix = contract_uri_wrap(i)
                for j in ns['nodeIds']:
                    if i != j:
                        jx = contract_uri_wrap(j)
                        digraph.add_edge(ix, jx, pred='equivalentTo')
    if logical_definitions is not None and 'logicalDefinitionAxioms' in og:
        for a in og['logicalDefinitionAxioms']:
            ld = LogicalDefinition(contract_uri_wrap(a['definedClassId']),
                                   [contract_uri_wrap(x) for x in a['genusIds']],
                                   [(contract_uri_wrap(x['propertyId']),
                                     contract_uri_wrap(x['fillerId'])) for x in a['restrictions'] if x is not None])
            logical_definitions.append(ld)
                                
        

def convert_json_file(obographfile, **args):
    """
    Return a networkx MultiDiGraph of the ontologies
    serialized as a json string

    """
    f = open(obographfile, 'r')
    jsonstr = f.read()
    f.close()
    return convert_json_object(json.loads(jsonstr), **args)

def convert_json_object(obographdoc, **args):
    """
    Return a networkx MultiDiGraph of the ontologies
    serialized as a json object

    """
    digraph = networkx.MultiDiGraph()
    xref_graph = networkx.Graph()
    logical_definitions = []
    for og in obographdoc['graphs']:
        add_obograph_digraph(og, digraph, xref_graph=xref_graph, logical_definitions=logical_definitions, **args)

    return {
        'graph': digraph,
        'xref_graph': xref_graph,
        'graphdoc': obographdoc,
        'logical_definitions': logical_definitions
        }





from ontobio.golr.golr_query import GolrSearchQuery

def test_search():
    q = GolrSearchQuery("abnormal")
    print("Q={}".format(q))
    params = q.solr_params()
    print("PARAMS={}".format(params))
    results = q.exec()
    print("RESULTS={}".format(results))
    docs = results['docs']
    for r in docs:
        print(str(r))
    assert len(docs) > 0

def test_search_go_all():
    q = GolrSearchQuery("transport*", is_go=True)
    print("Q={}".format(q))
    params = q.solr_params()
    print("PARAMS={}".format(params))
    results = q.exec()
    print("RESULTS={}".format(results))
    docs = results['docs']
    for r in docs:
        print(str(r))
    assert len(docs) > 0
    print(str(results['facet_counts']))

def test_search_go_ontol():
    q = GolrSearchQuery("transport*", category='ontology_class', is_go=True)
    print("Q={}".format(q))
    params = q.solr_params()
    print("PARAMS={}".format(params))
    results = q.exec()
    print("RESULTS={}".format(results))
    docs = results['docs']
    for r in docs:
        print(str(r))
    assert len(docs) > 0
    



from ontobio.ontol_factory import OntologyFactory
from ontobio.golr.golr_associations import search_associations, GolrFields, map2slim

M=GolrFields()

HUMAN_SHH = 'NCBIGene:6469'
HOLOPROSENCEPHALY = 'HP:0001360'
TWIST_ZFIN = 'ZFIN:ZDB-GENE-050417-357'
DVPF = 'GO:0009953'
SUBJECTS = [TWIST_ZFIN,'ZFIN:ZDB-GENE-040426-1432']
SLIM = ['GO:0001525','GO:0048731','GO:0005634','GO:0005794']

def test_map2slim_golr():
    results = map2slim(subjects=SUBJECTS,
                       slim=SLIM,
                       object_category='function')

    assert len(results) > 0
    n_found = 0
    for r in results:
        n_found = n_found+1
        print("Subject: {} Slim:{} Assocs:{}".format(r['subject'],r['slim'],len(r['assocs'])))
    assert n_found > 0

def test_map2slim_ont():
    ont = OntologyFactory().create('go')
    relations=['subClassOf', 'BFO:0000050']
    m = ont.create_slim_mapping(subset_nodes=SLIM, relations=relations)
    print(str(m))
    assert m['GO:0005798'] == ['GO:0005794']

    searchresult = search_associations(subjects=SUBJECTS,
                                       object_category='function')

    ok = False
    for a in searchresult['associations']:
        gene = a['subject']['id']
        cid = a['object']['id']
        mcids = m.get(cid)
        ## e.g. ZFIN:ZDB-GENE-050417-357 : GO:0060037 -> ['GO:0048731']
        print("{} : {} -> {}".format(gene, cid, mcids))
        if gene == 'ZFIN:ZDB-GENE-050417-357' and mcids == ['GO:0048731']:
            ok = True
    assert ok

    



import os
import re
import subprocess

import setuptools

directory = os.path.dirname(os.path.abspath(__file__))

# version
init_path = os.path.join(directory, 'ontobio', '__init__.py')
with open(init_path) as read_file:
    text = read_file.read()
pattern = re.compile(r"^__version__ = ['\"]([^'\"]*)['\"]", re.MULTILINE)
version = pattern.search(text).group(1)

# long_description
readme_path = os.path.join(directory, 'README.md')
try:
    # copied from dhimmel/obonet:
    # Try to create an reStructuredText long_description from README.md
    args = 'pandoc', '--from', 'markdown', '--to', 'rst', readme_path
    long_description = subprocess.check_output(args)
    long_description = long_description.decode()
except Exception as error:
    # Fallback to markdown (unformatted on PyPI) long_description
    print('README.md conversion to reStructuredText failed. Error:')
    print(error)
    with open(readme_path) as read_file:
        long_description = read_file.read()

setuptools.setup(
    name='ontobio',
    version=version,
    author='Chris Mungall',
    author_email='cmungall@gmail.com',
    url='https://github.com/biolink/ontobio',
    description='Library for working with OBO Library Ontologies and associations',
    long_description=long_description,
    license='BSD',
    #packages=['ontobio'],
    packages=setuptools.find_packages(),

    keywords='ontology graph obo owl sparql networkx network',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Visualization'
    ],

    # Dependencies
    install_requires=[
        'networkx',
        'pyyaml',
        'pysolr',
        'requests',
        'sparqlwrapper',
        'cachier',
        'prefixcommons',
        'marshmallow',
        'scipy'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['plotly'],
        'test': ['pytest'],
    },
    scripts=['bin/ogr.py', 'bin/ontobio-assoc.py', 'bin/ontobio-parse-assocs.py']
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    #entry_points={
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    #},
)



from ontobio.io.gafparser import GafParser, GpadParser
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory

POMBASE = "tests/resources/truncated-pombase.gaf"
POMBASE_GPAD = "tests/resources/truncated-pombase.gpad"
ONT = "tests/resources/go-truncated-pombase.json"


def test_skim():
    p = GafParser()
    results = p.skim(open(POMBASE,"r"))
    for r in results:
        print(str(r))
        (s,sn,o) = r
        assert o.startswith('GO:')
        assert s.startswith('PomBase:')
    
def test_parse_gaf():
    parse_with(POMBASE, GafParser())
def test_parse_gpad():
    parse_with(POMBASE_GPAD, GpadParser())
    
def parse_with(f, p):
    is_gaf = f == POMBASE
    ont = OntologyFactory().create(ONT)
    results = p.parse(open(f,"r"))
    r1 = results[0]
    assert r1['evidence']['with_support_from'] == ['SGD:S000001583']
    assert r1['evidence']['has_supporting_reference'] == ['GO_REF:0000024']

    if is_gaf:
        assert r1['subject']['label'] == 'ypf1'
    for r in results:
        #print(str(r))
        sid = r['subject']['id']
        prov = r['provided_by']
        assert prov == 'PomBase' or prov == 'UniProt'
        assert r['object']['id'].startswith('GO:')
        assert sid.startswith('PomBase:') or (not is_gaf and sid.startswith('PR'))
        if is_gaf:
            assert r['subject']['taxon']['id'] =='NCBITaxon:4896'

    for m in p.report.messages:
        print("MESSAGE: {}".format(m))
    assert len(p.report.messages) == 0
    print(p.report.to_markdown())
    
def test_validate_go_idspaces():
    ont = OntologyFactory().create(ONT)
    p = GafParser()
    p.config.class_idspaces = ['FOOZ']
    assocs = p.parse(open(POMBASE,"r"))
    for m in p.report.messages:
        print("MESSAGE: {}".format(m))
    assert len(assocs) == 0
    assert len(p.report.messages) > 1
    summary = p.report.to_report_json()['summary']
    assert summary['association_count'] == 0
    assert summary['line_count'] > 300
    print(p.report.to_markdown())
    



import logging
import yaml
from marshmallow import Schema, fields, pprint, post_load

class OntologyConfigSchema(Schema):
    """
    Set of ontologies
    """
    id = fields.Str(description="local identifier")
    handle = fields.Str(description="ontology handle")
    pre_load = fields.Bool(description="if true, load this ontology at startup")

class ConfigSchema(Schema):
    """
    Marshmallow schema for configuration objects.
    """
    sparql_url = fields.Url(description="SPARQL URL to use for ontology queries")
    solr_url = fields.Url()
    solr_search_url = fields.Url()
    go_solr_url = fields.Url()
    go_solr_search_url = fields.Url()
    scigraph_ontology_url = fields.Url()
    scigraph_data_url = fields.Url()
    ontologies = fields.List(fields.Nested(OntologyConfigSchema))
    
    @post_load
    def make_object(self, data):
        logging.info("POST-LOAD: {}".format(data))
        return Config(**data)

    
class Config():
    """
    A configuration object determines which external service URLs are used
    for different kinds of calls.

    """
    def __init__(self,
                 solr_assocs_url = "https://solr.monarchinitiative.org/solr/golr",
                 amigo_solr_assocs_url = "http://golr.berkeleybop.org",
                 solr_search_url = "https://solr-dev.monarchinitiative.org/solr/search",
                 amigo_solr_search_url = "http://golr.berkeleybop.org",
                 sparql_url = "http://sparql.hegroup.org/sparql",
                 scigraph_ontology_url = None,
                 scigraph_data_url = None,
                 ontologies = {},
                 use_amigo_for = "function"):
        self.solr_assocs_url = solr_assocs_url
        self.amigo_solr_assocs_url = amigo_solr_assocs_url
        self.solr_search_url = solr_assocs_url
        self.amigo_solr_search_url = amigo_solr_assocs_url
        self.sparql_url = sparql_url
        self.scigraph_ontology_url = scigraph_ontology_url
        self.scigraph_data_url = scigraph_data_url
        self.ontologies = ontologies
        self.use_amigo_for = use_amigo_for

    def get_solr_search_url(use_amigo=False):
        """
        Return solr URL to be used for lexical entity searches

        A solr search URL is used to search entities/concepts based on a limited set of parameters.

        Arguments
        ---------
        use_amigo : bool
            If true, get the URL for the GO/AmiGO instance of GOlr. This is typically used for category='function' queries
        """
        url = self.solr_search_url
        if use_amigo:
            url = self.amigo_solr_search_url
        return url
            
    def get_solr_assocs_url(use_amigo=False):
        """
        Return solr URL to be used for assocation (enhanced triple) queries

        A solr assocs URL is used to query triple-patterns in Solr, ie subject-relation-object

        There are two possible schemas: Monarch and AmiGO. The AmiGO schema is used for
        querying the GO and Planteome Golr instances
        """
        url = self.solr_assocs_url
        if use_amigo:
            url = self.amigo_solr_assocs_url
        return url

class Session():
    """
    Configuration for current session
    """
    default_config_path = 'conf/config.yaml'
    config = None

"""
    Current session
"""
session = Session()

def get_config():
    """
    Return configuration for current session.

    When called for the first time, this will create a config object, using
    whatever is the default load path to find the config yaml
    """
    if session.config is None:
        logging.info("LOADING FROM: {}".format(session.default_config_path))
        session.config = load_config(session.default_config_path)
    else:
        logging.info("Using pre-loaded object: {}".format(session.config))
    return session.config

def load_config(path):
    f = open(path,'r')
    obj = yaml.load(f)
    schema = ConfigSchema()
    config = schema.load(obj).data
    errs = schema.validate(obj)
    if len(errs) > 0:
        logging.error("ERRS: {}".format(errs))
        raise ValueError('Error loading '+path)
    #config = Config()
    return config
